/*
 * gaim - Blogger (xml-rpc) Protocol Plugin
 *
 * Copyright (C) 2003, Herman Bloggs <hermanator12002@yahoo.com>
 * GPL Software Supported by www.Madster.com
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include "internal.h"

#include "blogger.h"

#include "gtkimhtml.h"
#include "gtkutils.h"

#include "accountopt.h"
#include "debug.h"
#include "notify.h"
#include "prpl.h"
#include "proxy.h"
#include "stock.h"
#include "version.h"

#include "util.h"
#include "gtkdialogs.h"

#include "xmlrpc_http.h"
#include "xmlrpc.h"

/*
 *  DEFINES
 */

/* Blogger API app key for Gaim plugin */
/* Old Dev Key */
/*#define BLOGGER_API_KEY        "684762C16A49E144F4FDF4136F3AACA7BF5C0F"*/
#define BLOGGER_API_KEY        "093A39F03B761308133C12D941F9E5E9F918EDB2"

/* Default Host info */
#define BLOGGER_HOST_URI       "www.blogger.com/api/"
#define OLD_BLOGGER_HOST_URI   "api.blogger.com/api/"
#define BLOGGER_HOST_PORT      80

#define BLOGGER_TIMEOUT        300000 /* 5 minutes */

#define XML_VERSION            "<?xml version=\"1.0\"?>"
#define HTTP_VERSION           "HTTP/1.0"
#define USER_AGENT_NAME        "Gaim Blogger"

#define CRLF                   "\r\n"

/*
 *  MACROS
 */
#define CHECK_POINTER(func, ptr) \
{ \
	if(!ptr) \
		gaim_debug(GAIM_DEBUG_WARNING, "blogger", "%s: Unexpected NULL pointer\n", #func); \
}

#define GET_OPT(options, setting)    ((options)&setting)
#define SET_OPT(options, setting)    ((options)|=setting)
#define UNSET_OPT(options, setting)  ((options&setting) ? options^=setting : 0)
#define RESET_OPT(options)           (options=0)

/*
 *  DATA STRUCTS
 */

/* Data structs for passing data between GTK windows */

struct template_data {
	GaimConnection *gc;
	char* blog_name;
};

struct post_data {
	GaimConnection *gc;
	BlogPost *bpost;
	GtkTreeSelection *sel;
	const char* new_content;
};

struct gen_editor_data {
	GtkWidget *win;
	GtkTextBuffer *txtbuf;
	void* data;
};

/*
 *  LOCALS & GLOBALS
 */
static GaimPlugin *my_protocol = NULL;

	
/*
 * since the API is lacking a timezone
 * and appears to be using PST as of 01/15/03
 * we will default to PST w/ unknown Daylight Savings.
 * This may need to be fixed at some point if they fix the API
 *
 * @see #calculate_server_tz_offset()
 */
static long server_tz_offset = 0;

/*
 *  Forward Declarations
 */
static void blogger_callback(gpointer data, gint source, GaimInputCondition cond);
static void blogger_edit_post(GaimConnection *gc, BlogPost *bpost, const char* content, void *data);
static void blogger_delete_post(GaimConnection *gc, char* postid, GtkWidget* win);

/**
 **  Private Code
 **/

static void calculate_server_tz_offset() {
	struct tm *t;
	time_t tt = 0;
	long offset = 0;
		
	time(&tt);
	t = localtime(&tt);
	offset = mktime(t);
	t = gmtime(&tt);
	offset -= mktime(t);
	 
	/* set the global variable */
	server_tz_offset = (offset + 28800); /* 8 Hours (PST offset) */
}

/*
 *  Helpers
 */

/* Returned string needs to be freed */
static void set_pending(BloggerData *bd, 
			BloggerReq pend,
			UserBlog *blog,
			PendingOps pend_ops,
			char *request) {
	bd->pending = pend;
	bd->pending_blog = blog;
	bd->pending_ops = pend_ops;
	bd->send_data = request;
	bd->sd_length = strlen(request);
}

static void reset_pending(BloggerData *bd) {
	bd->pending = none;
	bd->pending_blog = NULL;
	bd->pending_ops = NOOP;
	if(bd->send_data)
		g_free(bd->send_data);
	bd->send_data = NULL;
	bd->cb_data = NULL;
	bd->sd_length = 0;
}


static char* get_host_from_uri(const char* uri) {
	char* buf = g_strdup(uri);

	return strtok(buf, "/");
}

static time_t convert_iso8601(const char* date) {
	char *year;
	char *month;
	char *day;
	char *HH;
	char *MM;
	char *SS;
	struct tm t;
	time_t tt=0;

	if(!date || !(strlen(date)==17)) {
		gaim_debug(GAIM_DEBUG_WARNING, "blogger", "convert_iso8601: Invalid argument\n");
		return 0;
	}

	year = g_strndup(date, 4);
	month = g_strndup(date+=4, 2);
	day = g_strndup(date+=2, 2);
	HH = g_strndup(date+=3, 2);
	MM = g_strndup(date+=3, 2);
	SS = g_strndup(date+=3, 2);
	t.tm_sec = atoi(SS);
	t.tm_min = atoi(MM);
	t.tm_hour = atoi(HH);
	t.tm_mday = atoi(day);
	t.tm_mon = atoi(month)-1;
	t.tm_year = (atoi(year)-1900);
	t.tm_isdst = -1;
	tt = mktime(&t);
	
	/* the server is in PST and doesn't provide a TZ for formatting the date, so we hack */
	tt += server_tz_offset;
	
	g_free(year);
	g_free(month);
	g_free(day);
	g_free(HH);
	g_free(MM);
	g_free(SS);
	return tt;
}

static void free_blog(UserBlog *blog) {
	g_free(blog);
}

static void free_bloglist(GList *bl) {
	GList *tmp=bl;

	while(tmp) {
		free_blog((UserBlog*)tmp->data);
		tmp=tmp->next;
	}
	g_list_free(bl);
}

static void free_post(BlogPost *bpost) {
	g_free(bpost->postid);
	g_free(bpost->lastModified);
	g_free(bpost->status);
	g_free(bpost->userid);
	g_free(bpost->content);
	g_free(bpost->authorName);
	g_free(bpost->title);
	g_free(bpost->dateCreated);
	g_free(bpost->postDate);
	g_free(bpost->url);
	g_free(bpost);
}

static void free_postlist(GList *pl) {
	GList *tmp=pl;

	while(tmp) {
		free_post((BlogPost*)tmp->data);
		tmp=tmp->next;
	}
	g_list_free(pl);
}

static char* get_txtbuf_content(GtkTextBuffer *txtbuf) {
	GtkTextIter start;
	GtkTextIter end;

	gtk_text_buffer_get_bounds(txtbuf, &start, &end);
	return gtk_text_buffer_get_text(txtbuf, &start, &end, FALSE);
}

static char* blogger_get_method_string(BloggerReq method) {
	switch(method) {
#define CASE(x) case x: return #x
		CASE(none);
		CASE(newPost);
		CASE(editPost);
		CASE(deletePost);
		CASE(getRecentPosts);
		CASE(getUsersBlogs);
		CASE(getUserInfo);
		CASE(getTemplate);
		CASE(setTemplate);
#undef CASE
	}
	return "";
}

static UserBlog *blogger_find_blog(BloggerData *bd, const char* who) {
	GList *l = SAFE_POINTER_2(bd,blog_list);
	UserBlog *blog;

	while(l) {
		blog = (UserBlog*)l->data;
		if(strcmp(blog->blogName, who)==0)
			return blog;
		l = l->next;
	}
	return NULL;
}

static void blogger_remove_invalid_blogs(GaimConnection *gc) {
	GaimBuddyList *blist = gaim_get_blist();
	GaimBlistNode *group;
	GaimBlistNode *buddy;
	GaimAccount *account = gaim_connection_get_account(gc);

	if (blist == NULL || account == NULL) {
		return;
	}

	group = blist->root;
	while (group) {
		buddy = group->child;
		while (buddy) {
			if (GAIM_BLIST_NODE_IS_BUDDY(buddy)) {
				if (account == ((GaimBuddy*)buddy)->account)
					if(!blogger_find_blog((BloggerData*)gc->proto_data, ((GaimBuddy*)buddy)->name))
						gaim_blist_remove_buddy((GaimBuddy*)buddy);
			}
			buddy = buddy->next;
		}
		group = group->next;
	}
}

static void blogger_fault(GaimConnection *gc, BloggerReq request, struct xmlrpc_fault *f, PendingOps ops) {
	struct xmlrpc_struct *s;
	char* error=NULL;
	int err_code=0;
	char buf[128];
	char* e1=NULL;
	char* e2=NULL;
	GtkWidget *d;

	gaim_debug(GAIM_DEBUG_WARNING, "blogger", "blogger_fault: Server returned fault for method %d\n", request);

	s = XMLRPC_STRUCT(SAFE_POINTER_3(f, value, data));
	CHECK_POINTER(blogger_fault,s);

	while(s) {
		if(!SAFE_POINTER_3(s, member, name) || 
		   !SAFE_POINTER_4(s, member, value, data)) {
			if(!s->next) 
				break;
			else {
				s=s->next;
				continue;
			}
		}
		if(strcmp(s->member->name, "faultCode")==0) {
			err_code = atoi((char*)(s->member->value->data));
		}
		else if(strcmp(s->member->name, "faultString")==0)
			error = (char*)(s->member->value->data);
		else
			gaim_debug(GAIM_DEBUG_WARNING, "blogger", "blogger_get_fault: Unexpected fault struct member name\n");
		s=s->next;
	}

	/* Get rid of the 2 java.lang.Exception strings */
	if(error && (e1=strstr(error, "java.lang.Exception: "))) {
		e1=strchr(e1, ' ')+1;
		if((e2=strstr(e1, "java.lang.Exception: ")))
			e1=e2;
		error=strchr(e1, ' ')+1;
	}

	if(request == getUsersBlogs) {
		/* If doing a Gaim Signon .. */
		if(GET_OPT(((BloggerData*)gc->proto_data)->pending_ops, GET_USERS_BLOGS_STARTUP)) {
			gaim_connection_error(gc, (error ? error : _("Blogger server returned an error.")));
			return;
		}
	}

	if(!GET_OPT(ops, GET_USERS_BLOGS_PERIODIC)) {
		sprintf(buf,_("Blogger fault code %d"), err_code);
		d=gaim_notify_error(NULL, NULL, (error ? error : _("Blogger server returned an error.")), buf);
		gtk_window_set_modal(GTK_WINDOW(d), TRUE);
	}
}

static struct xmlrpc_response* blogger_process_response(GaimConnection *gc, 
							BloggerReq method,
							const char* recvbuf,
							PendingOps ops) {
	struct xmlrpc_response *resp = NULL;

	if(!(resp=xmlrpc_parse(recvbuf))) {
		gaim_debug(GAIM_DEBUG_WARNING, "blogger", "blogger_process_response: Error parsing xmlrpc server response.\n");
		if(!GET_OPT(ops, GET_USERS_BLOGS_PERIODIC))
			gaim_notify_error(NULL, NULL, _("Error parsing blogger server response."), NULL);
		return NULL;
	}

	if(resp->type == fault) {
		blogger_fault(gc, method, XMLRPC_FAULT(resp->data), ops);
		xmlrpc_free_response(resp);
		return NULL;
	}
	return resp;
}

/*
 *  Blogger specific UI
 */
static void show_generic_editor(char* title, 
				char* sub_title, 
				char* b1_label, 
				char* b2_label, 
				void* b1_cb, 
				void* b2_cb,
				void* destroy_cb,
				char* editor_text,
				void* data,
				gboolean modal) {
	GtkWidget *win, *box, *bbox, *scroll, *view, *frame;
	GtkWidget *close_button;
	GtkWidget *edit_button;
	GtkTextBuffer *txtbuf;
	struct gen_editor_data *editor_data;

	gaim_debug(GAIM_DEBUG_INFO, "blogger", "show_generic_editor\n");

	editor_data = g_new0(struct gen_editor_data, 1);

	editor_data->win = win = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_window_set_default_size(GTK_WINDOW(win), 400, 300);
	if(modal)
		gtk_window_set_modal(GTK_WINDOW(win), TRUE);

	/* Set destroy callback */
	g_signal_connect(GTK_OBJECT(win), "destroy", G_CALLBACK(destroy_cb), editor_data);

	view = gtk_text_view_new();
	gtk_text_view_set_wrap_mode(GTK_TEXT_VIEW(view), GTK_WRAP_WORD);
	editor_data->txtbuf = txtbuf = gtk_text_view_get_buffer(GTK_TEXT_VIEW(view));
	if(editor_text)
		gtk_text_buffer_insert_at_cursor(txtbuf, editor_text, strlen(editor_text));

	/* Set specific usage data */
	editor_data->data = data;

	scroll = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scroll), 
				       GTK_POLICY_AUTOMATIC,
				       GTK_POLICY_AUTOMATIC);
	gtk_scrolled_window_set_shadow_type(GTK_SCROLLED_WINDOW(scroll),
					    GTK_SHADOW_IN);
	gtk_container_set_border_width(GTK_CONTAINER(scroll), 5);
	gtk_container_add(GTK_CONTAINER(scroll), view);
	gtk_widget_show(scroll);
	box = gtk_vbox_new(FALSE, 5);

	frame = gtk_frame_new(sub_title);

	gtk_widget_show(frame);
	gtk_box_pack_start(GTK_BOX(box), frame, TRUE, TRUE, 0);
	gtk_container_add(GTK_CONTAINER(frame), scroll);

	gtk_window_set_title(GTK_WINDOW(win), title);
	gtk_container_set_border_width(GTK_CONTAINER(win), 10);
	gtk_container_add(GTK_CONTAINER(win), box);

	bbox = gtk_hbox_new(FALSE, 0);
	gtk_box_pack_start(GTK_BOX(box), bbox, FALSE, FALSE, 0);

	close_button = gaim_pixbuf_button_from_stock(b2_label, GTK_STOCK_CLOSE, GAIM_BUTTON_HORIZONTAL);
	gtk_box_pack_end(GTK_BOX(bbox), close_button, FALSE, FALSE, 5);
	g_signal_connect(GTK_OBJECT(close_button), "clicked", G_CALLBACK(b2_cb), win);

	edit_button = gaim_pixbuf_button_from_stock(b1_label, GAIM_STOCK_SEND, GAIM_BUTTON_HORIZONTAL);
	gtk_box_pack_end(GTK_BOX(bbox), edit_button, FALSE, FALSE, 5);
	g_signal_connect(GTK_OBJECT(edit_button), "clicked", G_CALLBACK(b1_cb), editor_data);

	gtk_widget_show_all(win);
}

static void window_close_cb(GtkButton *button, GtkWidget *widget) {
	gaim_debug(GAIM_DEBUG_INFO, "blogger", "window_close_cb\n");
	gtk_widget_destroy(GTK_WIDGET(widget));
}

/* Edit Post Seclection */
static void edit_post_destroy_cb(GtkWidget *widget, struct gen_editor_data *ged) {
	gaim_debug(GAIM_DEBUG_INFO, "blogger", "edit_post_destroy_cb\n");
	/* template specific data */
	g_free(ged->data);
	g_free(ged);
}

static void edit_post_send_cb(GtkButton *button, struct gen_editor_data *ged) {
	struct post_data *pd = ged->data;
	const char *content;

	gaim_debug(GAIM_DEBUG_INFO, "blogger", "send_edited_post\n");

	content = get_txtbuf_content(ged->txtbuf);
	pd->new_content = content;

	blogger_edit_post(pd->gc, pd->bpost, content, ged);
}

static void edit_post(GtkButton *button, GtkTreeSelection *sel) {
	GValue val = { 0, };
	GtkTreeIter iter;
	GtkTreeModel *model;
	BlogPost *bpost;
	BloggerData *bd;
	struct post_data *pd;
	char *title;
	char *subtitle;

	gaim_debug(GAIM_DEBUG_INFO, "blogger", "edit_post\n");
	if(!gtk_tree_selection_get_selected(sel, &model, &iter))
		return;
	gtk_tree_model_get_value(model, &iter, 1, &val);
	bpost = g_value_get_pointer(&val);
	if(!bpost)
		return;

	bd = g_object_get_data(G_OBJECT(button), "post-blog-data");

	title = g_strdup_printf(_("Blog Post Editor - %s"), SAFE_POINTER_3(bpost,blog,blogName));
	subtitle = g_strdup_printf(_("Post ID: %s"), bpost->postid);

	/* Attach data we need to process an edit_post command */
	pd = g_new0(struct post_data, 1);
	pd->sel = sel;
	pd->gc = bd->gc;
	pd->bpost = bpost;

	show_generic_editor(title,
			    subtitle,
			    _("Update Posting"), 
			    _("Cancel"),
			    edit_post_send_cb, 
			    window_close_cb,
			    edit_post_destroy_cb,
			    bpost->content,
			    (void*)pd,
			    1);
	g_free(subtitle);
	g_free(title);
}

static void delete_post(GtkButton *button, GtkTreeSelection *sel) {
	GtkWidget *win;
	GtkTreeModel *model;
	GtkTreeIter iter;
	GValue val = { 0, };
	BlogPost *bpost;
	BloggerData *bd;

	gaim_debug(GAIM_DEBUG_INFO, "blogger", "delete_post\n");

	if(!gtk_tree_selection_get_selected(sel, &model, &iter))
		return;
	gtk_tree_model_get_value(model, &iter, 1, &val);
	bpost = g_value_get_pointer(&val);

	if(!bpost)
		return;
	win = g_object_get_data(G_OBJECT(button), "post-del-win");
	bd = g_object_get_data(G_OBJECT(button), "post-del-blogdata");
	blogger_delete_post(bd->gc, bpost->postid, win);
}

static void show_post_contents(BlogPost *bpost, GtkWidget *layout) {
	int options = GTK_IMHTML_NO_COMMENTS | GTK_IMHTML_NO_TITLE | GTK_IMHTML_NO_SCROLL;
	char *text, *date, *url;
	time_t t;
	
	gtk_imhtml_clear(GTK_IMHTML(layout));

	t = convert_iso8601(bpost->lastModified);
	date = ctime(&t);

	url = bpost->url ? gaim_markup_linkify(bpost->url) : g_strdup("");

	text = g_strdup_printf(
#if 0
			       "<b>%s </b>%s\n"        	/* Title */
#endif
			       "<b>%s </b>%s\n"      	/* Author */
			       "<b>%s </b>%s"		/* Last Modified */
			       "<b>%s </b>%s\n\n",   	/* URL */ 
#if 0
			       _("Post Title:"), (bpost->title ? bpost->title : ""),
#endif
			       _("Author:"), bpost->authorName,
			       _("Last Modified:"), date,
			       _("URL:"), url);

	g_free(url);

	gtk_imhtml_append_text(GTK_IMHTML(layout), text, options);
	gtk_imhtml_append_text(GTK_IMHTML(layout), bpost->content, options);
}

static void select_post(GtkTreeSelection *sel, GtkTreeModel *model) {
	GValue val = { 0, };
	GtkTreeIter iter;
	GtkWidget *layout;

	if(!gtk_tree_selection_get_selected(sel, &model, &iter))
		return;

	layout = g_object_get_data(G_OBJECT(model), "post-display");

	gtk_tree_model_get_value(model, &iter, 1, &val);
	show_post_contents(g_value_get_pointer(&val), layout);
}

static void destroy_post_selector(GtkWidget *window, gpointer data) {
	free_postlist((GList*)data);
	gtk_widget_destroy(window);
}

void show_post_selector(BloggerData *bd, GList *posts) {
	GtkWidget *window;
	GtkWidget *box;
	GtkWidget *hbox;
	GtkWidget *bbox;
	GtkWidget *sw;
	GtkWidget *layout;
	GtkWidget *close_button, *edit_button, *delete_button;
	GtkListStore *list_store;
	GtkWidget *tree_view;
	GtkTreeSelection *sel = NULL;
	GtkTreePath *path;
	GtkWidget *frame;
	gchar buf[256];
	GList *L = posts;

	char convo_start[64];

	/* Set up window */
	window = gtk_window_new(GTK_WINDOW_TOPLEVEL);

	/* Free posts list when window is destroyed */
	g_signal_connect(GTK_OBJECT(window), "destroy", G_CALLBACK(destroy_post_selector), posts);

	g_snprintf(buf, sizeof buf, _("Edit recent blog posts"));
	gtk_window_set_title(GTK_WINDOW(window), buf);
	gtk_container_set_border_width(GTK_CONTAINER(window), 10);
	gtk_window_set_resizable(GTK_WINDOW(window), TRUE);
/*	***I can't imagine why this would be desirable for this window
	gtk_window_set_modal(GTK_WINDOW(window), TRUE);
*/
	gtk_widget_realize(window);

	layout = gtk_imhtml_new(NULL, NULL);
	bbox = gtk_hbox_new(FALSE, 0);

	box = gtk_vbox_new(FALSE, 5);
	gtk_container_add(GTK_CONTAINER(window), box);

	hbox = gtk_hbox_new(FALSE, 5);
	gtk_box_pack_start(GTK_BOX(box), hbox, TRUE, TRUE, 0);

	/* Set up postid list */
	list_store = gtk_list_store_new(2,
					G_TYPE_STRING,
					G_TYPE_POINTER);

	/* Attach imhtml widget to the tree model .. used when a list item is selcted */
	g_object_set_data(G_OBJECT(GTK_TREE_MODEL(list_store)), "post-display", layout);

	tree_view = gtk_tree_view_new_with_model(GTK_TREE_MODEL(list_store));
	
	gtk_tree_view_set_headers_visible(GTK_TREE_VIEW(tree_view), FALSE);
	
	gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW(tree_view),
						    -1, "", gtk_cell_renderer_text_new(), "text", 0, NULL);
	
	sel = gtk_tree_view_get_selection(GTK_TREE_VIEW(tree_view));
	g_signal_connect(G_OBJECT(sel), "changed",
			 G_CALLBACK(select_post),
			 NULL);
	
	frame = gtk_frame_new(_("Recent posts"));
	gtk_widget_show(frame);
	
	sw = gtk_scrolled_window_new(NULL, NULL);
	gtk_container_set_border_width(GTK_CONTAINER(sw), 5);
	gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(sw), tree_view);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(sw),
				       GTK_POLICY_AUTOMATIC, GTK_POLICY_ALWAYS);
	gtk_widget_set_size_request(sw, 220, 220);
	gtk_container_add(GTK_CONTAINER(frame), sw);
	gtk_box_pack_start(GTK_BOX(hbox), frame, TRUE, TRUE, 0);
	
	/* Loop through Posts.. */
	while (L) {
		BlogPost *bpost = (BlogPost*)L->data;
		long t_time;
		char tbuf[25];
		GtkTreeIter iter;
		
		t_time = convert_iso8601(bpost->postDate);
		strftime(tbuf, sizeof(tbuf), "%a %b %d", gmtime(&t_time));

		g_snprintf(convo_start, sizeof(convo_start), "%s - ID: %s", 
			   tbuf, bpost->postid);

		gtk_list_store_append(list_store, &iter);
		/* Attach blog post to list */
		gtk_list_store_set(list_store, &iter,
				   0, convo_start,
				   1, bpost,
				   -1);

		L = L->next;
	}/* end while */

	/* Select first item in the list */
	path = gtk_tree_path_new_first();
	gtk_tree_selection_select_path(sel, path);
	gtk_tree_path_free(path);
	
	g_object_unref(G_OBJECT(list_store));

	/* Blog Post Display */
	frame = gtk_frame_new(_("Post details"));
	gtk_widget_show(frame);

	sw = gtk_scrolled_window_new(NULL, NULL);
	gtk_container_set_border_width(GTK_CONTAINER(sw), 5);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(sw), GTK_POLICY_AUTOMATIC, GTK_POLICY_ALWAYS);
	gtk_scrolled_window_set_shadow_type(GTK_SCROLLED_WINDOW(sw), GTK_SHADOW_IN);
	gtk_widget_set_size_request(sw, 390, 220);
	gtk_container_add(GTK_CONTAINER(frame), sw);
	gtk_box_pack_start(GTK_BOX(hbox), frame, TRUE, TRUE, 0);

	g_signal_connect(GTK_OBJECT(layout), "url_clicked", G_CALLBACK(gaim_notify_uri), NULL);
	gtk_container_add(GTK_CONTAINER(sw), layout);
	gaim_setup_imhtml(layout);

	gtk_box_pack_start(GTK_BOX(box), bbox, FALSE, FALSE, 0);
	gtk_widget_set_sensitive(bbox, FALSE);

	close_button = gaim_pixbuf_button_from_stock(_("Close"), GTK_STOCK_CLOSE, GAIM_BUTTON_HORIZONTAL);
	gtk_box_pack_end(GTK_BOX(bbox), close_button, FALSE, FALSE, 5);
	g_signal_connect(GTK_OBJECT(close_button), "clicked", G_CALLBACK(window_close_cb), window);

	edit_button = gaim_pixbuf_button_from_stock(_("Edit"), GTK_STOCK_PREFERENCES, GAIM_BUTTON_HORIZONTAL);
	gtk_box_pack_end(GTK_BOX(bbox), edit_button, FALSE, FALSE, 5);
	g_signal_connect(GTK_OBJECT(edit_button), "clicked", G_CALLBACK(edit_post), sel);
	g_object_set_data(G_OBJECT(edit_button), "post-blog-data", bd);

	delete_button = gaim_pixbuf_button_from_stock(_("Delete"), GTK_STOCK_REMOVE, GAIM_BUTTON_HORIZONTAL);
	gtk_box_pack_end(GTK_BOX(bbox), delete_button, FALSE, FALSE, 5);
	g_signal_connect(GTK_OBJECT(delete_button), "clicked", G_CALLBACK(delete_post), sel);
	g_object_set_data(G_OBJECT(delete_button), "post-del-win", window);
	g_object_set_data(G_OBJECT(delete_button), "post-del-blogdata", bd);

	gtk_widget_show_all(window);
	gtk_widget_set_sensitive(bbox, TRUE);
	return;
}

/*
 *  Blogger Method Send Routines
 */

static char* print_xmlrpc_param(char* buffer, xmlrpc_type type, char* param) {
	char* tmp;

	switch(type) {
	case Integer_T:
		tmp = g_strdup_printf("      <param><value><int>%s</int></value></param>" CRLF, param);
		break;
	case Boolean_T:
		tmp = g_strdup_printf("      <param><value><boolean>%s</boolean></value></param>" CRLF, param);
		break;
	case String_T: {
		char* tmp1 = g_markup_escape_text(param, strlen(param));
		tmp = g_strdup_printf("      <param><value><string>%s</string></value></param>" CRLF, tmp1);
		g_free(tmp1);
		break;
	}
	default:
		gaim_debug(GAIM_DEBUG_WARNING, "blogger", "print_xmlrpc_param: implementation missing for type %d\n", type);
		return NULL;
	}

	buffer = (char*)g_realloc(buffer, strlen(buffer) + strlen(tmp) + 1);
	strcat(buffer, tmp);
	g_free(tmp);
	return buffer;
}

static char* blogger_xmlrpc_method(BloggerReq method, ...) {
	va_list ap;
	gchar *params[10];
	int pcount=0;
	char *buffer=NULL;
	char *tmp;
	/* number of params for each method - should be in same order as in enum declaration */
	int args[] = {0,        /* none - not a method */
		      6,        /* newPost */
		      6,        /* editPost */
		      5,        /* deletePost */
		      5,        /* getRecentPosts */
		      3,        /* getUsersBlogs */
		      3,        /* getUserInfo */
		      5,        /* getTemplate */
		      6};       /* setTemplate */

	va_start(ap, method);
	while((params[pcount]=va_arg(ap, char*)))
		pcount++;
	va_end(ap);

	if(pcount!=args[method]) {
		gaim_debug(GAIM_DEBUG_WARNING, "blogger", "xmlrpc_xml_method: Error - incorrect number of params\n");
		return NULL;
	}

	buffer = g_strdup_printf(XML_VERSION CRLF
				 "<methodCall>" CRLF
				 "   <methodName>blogger.%s</methodName>" CRLF
				 "   <params>" CRLF,
				 blogger_get_method_string(method));

	switch(method) {
	/* 3 string params */
	case getUsersBlogs:
	case getUserInfo:
		buffer=print_xmlrpc_param(buffer, String_T, params[0]);
		buffer=print_xmlrpc_param(buffer, String_T, params[1]);
		buffer=print_xmlrpc_param(buffer, String_T, params[2]);
		break;
	/* 4 string params 1 int */
	case getRecentPosts:
		buffer=print_xmlrpc_param(buffer, String_T, params[0]);
		buffer=print_xmlrpc_param(buffer, String_T, params[1]);
		buffer=print_xmlrpc_param(buffer, String_T, params[2]);
		buffer=print_xmlrpc_param(buffer, String_T, params[3]);
		buffer=print_xmlrpc_param(buffer, Integer_T, params[4]);
		break;
	/* 4 string params 1 bool */
	case deletePost:
		buffer=print_xmlrpc_param(buffer, String_T, params[0]);
		buffer=print_xmlrpc_param(buffer, String_T, params[1]);
		buffer=print_xmlrpc_param(buffer, String_T, params[2]);
		buffer=print_xmlrpc_param(buffer, String_T, params[3]);
		buffer=print_xmlrpc_param(buffer, Boolean_T, params[4]);
		break;
	/* 5 string params */
	case getTemplate:
		buffer=print_xmlrpc_param(buffer, String_T, params[0]);
		buffer=print_xmlrpc_param(buffer, String_T, params[1]);
		buffer=print_xmlrpc_param(buffer, String_T, params[2]);
		buffer=print_xmlrpc_param(buffer, String_T, params[3]);
		buffer=print_xmlrpc_param(buffer, String_T, params[4]);
		break;
	/* 5 string 1 bool params */
	case editPost:
	case newPost:
		buffer=print_xmlrpc_param(buffer, String_T, params[0]);
		buffer=print_xmlrpc_param(buffer, String_T, params[1]);
		buffer=print_xmlrpc_param(buffer, String_T, params[2]);
		buffer=print_xmlrpc_param(buffer, String_T, params[3]);
		buffer=print_xmlrpc_param(buffer, String_T, params[4]);
		buffer=print_xmlrpc_param(buffer, Boolean_T, params[5]);
		break;
	/* 6 string params */
	case setTemplate:
		buffer=print_xmlrpc_param(buffer, String_T, params[0]);
		buffer=print_xmlrpc_param(buffer, String_T, params[1]);
		buffer=print_xmlrpc_param(buffer, String_T, params[2]);
		buffer=print_xmlrpc_param(buffer, String_T, params[3]);
		buffer=print_xmlrpc_param(buffer, String_T, params[4]);
		buffer=print_xmlrpc_param(buffer, String_T, params[5]);
	case none:
		break;
	}
	
	tmp=g_strdup_printf("   </params>" CRLF
			    "</methodCall>");
	buffer = (char*)g_realloc(buffer, strlen(buffer) + strlen(tmp) + 1);
	strcat(buffer, tmp);
	g_free(tmp);
	return buffer;
}

static void blogger_connect(gpointer data, gint source, GaimInputCondition cond) {
	GaimConnection *gc = data;
	BloggerData *bd;

        gaim_debug(GAIM_DEBUG_INFO, "blogger", "blogger_connect\n");

	if (!g_list_find(gaim_connections_get_all(), gc)) {
		close(source);
		return;
	}

	bd = gc->proto_data;

	if(!source) {
		char buf[128];
		snprintf(buf, sizeof buf, _("Connection error: %s."), strerror(errno));
		gaim_notify_error(NULL, NULL, buf, NULL);
	}
	/* Just completed proxy_connect, ready to send data to the server */
	else if(bd->send_data && bd->sd_length > 0) {
		gaim_debug(GAIM_DEBUG_INFO, "blogger", "blogger_connect: Sending Blogger Request:\n\n%s\n\n", bd->send_data);
		if(xmlrpc_http_post(source,
				    "Gaim Blogger Plugin v" VERSION,
				    "text/xml",
				    gaim_account_get_string(gc->account, "server", BLOGGER_HOST_URI),
				    bd->send_data, 
				    bd->sd_length))
			gc->inpa = gaim_input_add(source, GAIM_INPUT_READ, blogger_callback, gc);
		else {
			gaim_notify_error(NULL, NULL, _("Error sending blogger request."), NULL);
			close(source);
		}
	}
	else {
		gaim_debug(GAIM_DEBUG_INFO, "blogger", "blogger_connect: Nothing to send\n");
		close(source);
	}
	/* Clear send buffer */
	g_free(bd->send_data);
	bd->send_data = NULL;
	bd->sd_length = 0;
}

static void blogger_send_request(GaimConnection *gc, 
				 BloggerReq pend, 
				 char* request, 
				 UserBlog* blog, 
				 PendingOps options) {
	BloggerData *bd = gc->proto_data;
	char* host;

	/* If busy, add request to request queue */
	if(bd->pending != none) {
		BloggerRequest *req = g_new0(BloggerRequest, 1);
		req->pend = pend;
		req->request = request;
		req->blog = blog;
		req->options = options;
		g_queue_push_head(bd->request_queue, req);
		return;
	}
	set_pending(bd, pend, blog, options, request);

	host = get_host_from_uri(gaim_account_get_string(gc->account, "server", BLOGGER_HOST_URI));

	if (gaim_proxy_connect(gc->account, 
			       host,
			       gaim_account_get_int(gc->account, "port", BLOGGER_HOST_PORT),
			       blogger_connect, gc) < 0) {
		gaim_debug(GAIM_DEBUG_WARNING, "blogger", "Error connecting to server\n");
		reset_pending(bd);
	}
	g_free(host);
}

/*
 * Blogger XMLRPC Methods
 */

static void blogger_get_users_blogs(GaimConnection *gc, PendingOps options) {
	char* xml_method;

	gaim_debug(GAIM_DEBUG_INFO, "blogger", "blogger_get_users_blogs\n");

	if(!(xml_method=blogger_xmlrpc_method(getUsersBlogs,
					      BLOGGER_API_KEY,
					      gaim_account_get_username(gc->account),
					      gaim_account_get_password(gc->account),
					      NULL)))
		return;
        gaim_debug(GAIM_DEBUG_INFO, "blogger", "done calling blogger_xmlrpc_method\n");

	blogger_send_request(gc, getUsersBlogs, xml_method, NULL, options);
}

static void blogger_new_post(GaimConnection *gc, const char* blog_name, const char* content) {
	char* xml_method;
	BloggerData* bd = gc->proto_data;
	UserBlog* blog;

	gaim_debug(GAIM_DEBUG_INFO, "blogger", "blogger_new_post\n");
	if(!blog_name) {
		return;
	}
	if(!(blog=blogger_find_blog(bd, blog_name))) {
		gaim_debug(GAIM_DEBUG_WARNING, "blogger", "blogger_new_post: Couldn't find blog for name: %s\n", blog_name);
		return;
	}

	if(!(xml_method=blogger_xmlrpc_method(newPost,
					      BLOGGER_API_KEY,
					      blog->blogid,
					      gaim_account_get_username(gc->account),
					      gaim_account_get_password(gc->account),
					      content,
					      "1",
					      NULL)))
		return;
	blogger_send_request(gc, newPost, xml_method, blog, 0);
}

static void blogger_edit_post(GaimConnection *gc, BlogPost *bpost, const char* content, void *data) {
	char* xml_method;
	BloggerData *bd = gc->proto_data;

	if(!(xml_method=blogger_xmlrpc_method(editPost,
					      BLOGGER_API_KEY,
					      bpost->postid,
					      gaim_account_get_username(gc->account),
					      gaim_account_get_password(gc->account),
					      content,
					      "1",
					      NULL)))
		return;
	bd->cb_data = data;
	blogger_send_request(gc, editPost, xml_method, bpost->blog, 0);
	/* Update last published .. */
	blogger_get_users_blogs(gc, 0);
}

static void blogger_delete_post(GaimConnection *gc, char* postid, GtkWidget* win) {
	char* xml_method;
	BloggerData *bd = gc->proto_data;

	if(!postid)
		return;

	if(!(xml_method=blogger_xmlrpc_method(deletePost,
					      BLOGGER_API_KEY,
					      postid,
					      gaim_account_get_username(gc->account),
					      gaim_account_get_password(gc->account),
					      "1",
					      NULL)))
		return;
	bd->cb_data = win;
	blogger_send_request(gc, deletePost, xml_method, NULL, 0);
	/* Update last published .. */
	blogger_get_users_blogs(gc, 0);
}

static void blogger_get_recent_posts(GaimConnection *gc, const char* blog_name, int postnum, PendingOps options) {
	char* xml_method;
	BloggerData *bd = gc->proto_data;
	UserBlog* blog;
	char buf[10];

	gaim_debug(GAIM_DEBUG_INFO, "blogger", "blogger_get_recent_posts\n");
	if(!blog_name) {
		return;
	}
	if(!(blog=blogger_find_blog(bd, blog_name))) {
		gaim_debug(GAIM_DEBUG_WARNING, "blogger", "blogger_get_recent_posts: Couldn't find blog for name: %s\n", blog_name);
		return;
	}
	snprintf(buf, sizeof buf, "%d", postnum);
	if(!(xml_method=blogger_xmlrpc_method(getRecentPosts,
					      BLOGGER_API_KEY,
					      blog->blogid,
					      gaim_account_get_username(gc->account),
					      gaim_account_get_password(gc->account),
					      buf,
					      NULL)))
		return;
	
	blogger_send_request(gc, getRecentPosts, xml_method, blog, options);
}

static void blogger_get_template(GaimConnection *gc, const char* blog_name) {
	char* xml_method;
	BloggerData* bd = gc->proto_data;
	UserBlog* blog;

	gaim_debug(GAIM_DEBUG_INFO, "blogger", "blogger_get_template\n");
	if(!blog_name) {
		return;
	}
	if(!(blog=blogger_find_blog(bd, blog_name))) {
		gaim_debug(GAIM_DEBUG_INFO, "blogger", "blogger_get_template: Couldn't find blog for name: %s\n", blog_name);
		return;
	}

	if(!(xml_method=blogger_xmlrpc_method(getTemplate,
					      BLOGGER_API_KEY,
					      blog->blogid,
					      gaim_account_get_username(gc->account),
					      gaim_account_get_password(gc->account),
					      "main",
					      NULL)))
		return;
	
	blogger_send_request(gc, getTemplate, xml_method, blog, 0);
}

static void blogger_set_template(GaimConnection *gc, char* blog_name, char* template, void* data) {
	char* xml_method;
	BloggerData *bd = gc->proto_data;
	UserBlog* blog;

	gaim_debug(GAIM_DEBUG_INFO, "blogger", "blogger_set_template\n");
	if(!blog_name) {
		return;
	}
	if(!(blog=blogger_find_blog(bd, blog_name))) {
		gaim_debug(GAIM_DEBUG_WARNING, "blogger", "blogger_set_template: Couldn't find blog for name: %s\n", blog_name);
		return;
	}

	if(!(xml_method=blogger_xmlrpc_method(setTemplate,
					      BLOGGER_API_KEY,
					      blog->blogid,
					      gaim_account_get_username(gc->account),
					      gaim_account_get_password(gc->account),
					      template,
					      "main",
					      NULL)))
		return;
	
	bd->cb_data = data;
	blogger_send_request(gc, setTemplate, xml_method, blog, 0);
	/* Update last published .. */
	blogger_get_users_blogs(gc, 0);
}

/*
 * Blogger Callbacks
 */

/* Blogger XMLRPC Method Callbacks */
static void blogger_new_post_cb(GaimConnection *gc, const char* recvbuf, UserBlog *blog) {
	GaimConversation *cnv;
	struct xmlrpc_response *resp;
	char* post_id;
	char *tmpbuf=NULL;

	gaim_debug(GAIM_DEBUG_INFO, "blogger", "blogger_new_post_cb\n");
	cnv = gaim_find_conversation(blog->blogName);

	if(!(resp = blogger_process_response(gc, newPost, recvbuf, 0))) {
		gaim_conversation_write(cnv, NULL, _("Blogger Post Failed."), GAIM_MESSAGE_SYSTEM, time(NULL));
		gaim_debug(GAIM_DEBUG_WARNING, "blogger", "Blogger post Failed.\n");
		return;
	}

	/* On Success we get a Blog Post ID */
	post_id = (char*)SAFE_POINTER_3(XMLRPC_PARAM(SAFE_POINTER_2(resp,data)),value,data);

	/* reset our last posted */
	if(blog && post_id)
		strcpy(blog->our_last_postid, post_id);

	tmpbuf = g_strdup_printf("%s%s%s",_("Blogger Post Success (Post ID: "), post_id ? post_id : "None", ").");
	gaim_debug(GAIM_DEBUG_INFO, "blogger", "Printing to gtkimhtml:\n%s\n", tmpbuf);

	if(cnv)
		gaim_conversation_write(cnv, NULL, tmpbuf, GAIM_MESSAGE_SYSTEM, time(NULL));
	else
		gaim_notify_info(NULL, NULL, tmpbuf, NULL);
	g_free(tmpbuf);

	gaim_debug(GAIM_DEBUG_INFO, "blogger", "Blogger post success.\n");
	xmlrpc_free_response(resp);

	/* Refresh list after Post */
	blogger_get_users_blogs(gc, 0);
}

static void blogger_edit_post_cb(GaimConnection *gc, const char* recvbuf, UserBlog *blog, struct gen_editor_data* ged) {
	struct xmlrpc_response *resp;
	struct post_data *pd = SAFE_POINTER_2(ged,data);
	char* ret;

	gaim_debug(GAIM_DEBUG_INFO, "blogger", "blogger_edit_post_cb\n");

	if(!(resp = blogger_process_response(gc, editPost, recvbuf, 0)))
		return;

	ret = (char*)SAFE_POINTER_3(XMLRPC_PARAM(SAFE_POINTER_2(resp,data)),value,data);
	if(atoi(ret) == 1) {
		gaim_debug(GAIM_DEBUG_INFO, "blogger", "blogger_edit_post_cb: Success\n");
		/* Close Edit window and update Blog editor's imhtml window with new post */
		if(pd) {
			GtkTreeModel *model=NULL;
			g_free(pd->bpost->content);
			pd->bpost->content = g_strdup(pd->new_content);
			select_post(pd->sel, model);
			gtk_widget_destroy(ged->win);
		}
	}
	else {
		gaim_debug(GAIM_DEBUG_WARNING, "blogger", "blogger_edit_post_cb: Return without error, but response unexpected\n");
		/* display error dialog */
	}

	xmlrpc_free_response(resp);
}

static void blogger_delete_post_cb(GaimConnection *gc, const char* recvbuf, GtkWidget* win) {
	struct xmlrpc_response *resp;
	char* ret;

        gaim_debug(GAIM_DEBUG_INFO, "blogger", "blogger_delete_post_cb\n");

	if(!(resp = blogger_process_response(gc, deletePost, recvbuf, 0)))
		return;
	
	ret = (char*)SAFE_POINTER_3(XMLRPC_PARAM(SAFE_POINTER_2(resp,data)),value,data);
	if(atoi(ret) == 1) {
		gaim_debug(GAIM_DEBUG_INFO, "blogger", "blogger_delete_post_cb: Success\n");
		gtk_widget_destroy(win);
	}
	else
		gaim_debug(GAIM_DEBUG_WARNING, "blogger", "blogger_delete_post_cb: Return without error, but response unexpected: %s\n", ret);
	xmlrpc_free_response(resp);
}

static void blogger_get_recent_posts_cb(GaimConnection *gc, const char* recvbuf, UserBlog *blog, PendingOps ops) {
	BloggerData *bd = gc->proto_data;
	struct xmlrpc_response *resp;
	struct xmlrpc_data *data;
	BlogPost *bpost=NULL;
	GList *post_list=NULL;

	gaim_debug(GAIM_DEBUG_INFO, "blogger", "blogger_get_recent_posts_cb\n");

	if(!(resp = blogger_process_response(gc, getRecentPosts, recvbuf, ops)))
		return;

	data = SAFE_POINTER_2(XMLRPC_ARRAY(SAFE_POINTER_3(XMLRPC_PARAM(SAFE_POINTER_2(resp,data)),value,data)),data);
	CHECK_POINTER(blogger_get_recent_posts_cb, data);

	/* Cycle through post structs in the array */
	while(data) {
		struct xmlrpc_struct *s = XMLRPC_STRUCT(SAFE_POINTER_3(data,value,data));
		CHECK_POINTER(blogger_get_recent_posts_cb, s);
		if(!s) {
			data = data->next;
			continue;
		}
		bpost = g_new0(BlogPost, 1);

		/* Cycle through struct members */
		while(s) {
			char* name = SAFE_POINTER_3(s,member,name);
			char* value = SAFE_POINTER_4(s,member,value,data);
			int len = s->member->value->data_len;

			if(!name) {
				gaim_debug(GAIM_DEBUG_WARNING, "blogger", "blogger_get_recent_posts_cb: Unexpected NULL pointer\n");
				s=s->next;
				continue;
			}

			if(strcmp(name, "postid")==0)
				 bpost->postid = g_strndup(value, len);
			else if(strcmp(name, "lastModified")==0)
				 bpost->lastModified = g_strndup(value, len);
			else if(strcmp(name, "status")==0)
				 bpost->status = g_strndup(value, len);
			else if(strcmp(name, "userid")==0)
				 bpost->userid = g_strndup(value, len);
			else if(strcmp(name, "content")==0)
				 bpost->content = g_strndup(value, len);
			else if(strcmp(name, "authorName")==0)
				 bpost->authorName = g_strndup(value, len);
			else if(strcmp(name, "title")==0)
				 bpost->title = g_strndup(value, len);
			else if(strcmp(name, "dateCreated")==0)
				 bpost->dateCreated = g_strndup(value, len);
			else if(strcmp(name, "postDate")==0)
				 bpost->postDate = g_strndup(value, len);
			else if(strcmp(name, "url")==0)
				 bpost->url = g_strndup(value, len);
			else {
				gaim_debug(GAIM_DEBUG_WARNING, "blogger", "blogger_get_recent_posts_cb: Unrecognized recent post " 
					     "struct member: %s\n", name);
			}
			s = s->next;
		}/* end while */
		bpost->blog = blog;
		post_list = g_list_append(post_list, bpost);
		data = data->next;
	}/* end while */

	xmlrpc_free_response(resp);

	/* What to do with the new post list ?*/
	if(post_list) {
		if(GET_OPT(ops, GET_POSTS_FOR_EDIT))
			show_post_selector(bd, post_list);
		else {
			if(GET_OPT(ops, GET_POSTS_CHECK_FOR_NEW_POST)) {
				if(bpost->postid && strcmp(blog->last_postid, bpost->postid)!=0) {
					if(bpost->postDate && strcmp(blog->last_postid_date, bpost->postDate) < 0) {
						SET_OPT(blog->blog_status, BLOGGER_NEW_POST);
						gaim_blist_update_buddy_status(blog->buddy, blog->buddy->uc);
					}
				}
				free_postlist(post_list);
			}
			if(GET_OPT(ops, GET_POSTS_LAST_POSTED)) {
				/* Set last_postid and last_postid_date */
				strcpy(blog->last_postid, bpost->postid ? bpost->postid : "");
				strcpy(blog->last_postid_date, bpost->postDate ? bpost->postDate : "");
				free_postlist(post_list);
			}
		}
	}
	/* This blog has no posts.. notify user */
        else if(GET_OPT(ops, GET_POSTS_FOR_EDIT)) {
		char buf[128];
		snprintf(buf, sizeof buf, _("There are no blog posts for '%s'"), blog->blogName);
		gaim_notify_info(NULL, NULL, buf, NULL);
	}
}

static void blogger_get_users_blogs_cb(GaimConnection *gc, const char* recvbuf, PendingOps ops) {
	struct xmlrpc_response *resp;
	struct xmlrpc_data *data;
	BloggerData *bd = gc->proto_data;
	char *gname; 

	gaim_debug(GAIM_DEBUG_INFO, "blogger", "blogger_get_users_blogs_cb\nrecvbuf:\n%s\n", recvbuf);

	if(GET_OPT(ops, GET_USERS_BLOGS_STARTUP)) {
		char buf[128];
		g_snprintf(buf, sizeof(buf), _("Retrieving Blog list."));
		gaim_connection_update_progress(gc, buf, 4, 5);
	}

	if(!(resp = blogger_process_response(gc, getUsersBlogs, recvbuf, ops))) {
		if(GET_OPT(ops, GET_USERS_BLOGS_STARTUP)) {
			gaim_connection_error(gc, _("Blogger plugin error"));
		}
		return;
	}

	data = SAFE_POINTER_2(XMLRPC_ARRAY(SAFE_POINTER_3(XMLRPC_PARAM(SAFE_POINTER_2(resp,data)),value,data)),data);

	CHECK_POINTER(blogger_get_users_blogs_cb, data);

	/* Group where these blogs should go */
	gname = g_strdup_printf("%s Blogs", gaim_account_get_username(gc->account));
	
	while(data) {
		struct xmlrpc_struct *s = XMLRPC_STRUCT(SAFE_POINTER_3(data,value,data));
		UserBlog *blog = NULL;
		gboolean isAdmin = FALSE;
		char* url = NULL;
		char* blogid = NULL;
		char* lastPublished = NULL;
		gboolean showTitle = FALSE;
		char* blogName = NULL;
		GaimBuddy *buddy = NULL;
		gboolean blog_changed = FALSE;

		CHECK_POINTER(blogger_get_users_blogs_cb, s);

		/* Loop through Blog struct members */
		while(s) {
			char* name = SAFE_POINTER_3(s,member,name);
			char* value = SAFE_POINTER_4(s,member,value,data);
			int len = s->member->value->data_len;

			if(!name) {
				gaim_debug(GAIM_DEBUG_WARNING, "blogger", "blogger_get_users_blogs_cb: Unexpected NULL pointer\n");
				s=s->next;
				continue;
			}
			if(!value)
				value = "";

			if(strcmp(name, "isAdmin")==0)
				isAdmin = atoi(value) ? TRUE : FALSE;
			else if(strcmp(name, "url")==0)
				url = g_strndup(value, len);
			else if(strcmp(name, "blogid")==0)
				blogid = g_strndup(value, len);
			else if(strcmp(name, "lastPublished")==0)
				lastPublished = g_strndup(value, len);
			else if(strcmp(name, "showTitle")==0)
				showTitle = atoi(value) ? TRUE : FALSE;
			else if(strcmp(name, "blogName")==0)
				blogName = g_strndup(value, len);
			else {
				gaim_debug(GAIM_DEBUG_WARNING, "blogger", "blogger_get_users_blogs_cb: Unrecognized user " 
					     "blog struct member: %s\n", name);
			}
			s=s->next;
		}/* end while */

		/* See if blog already exists */
		blog = blogger_find_blog(bd, blogName);

		if(!blog) {
			blog = g_new0(UserBlog, 1);
			bd->blog_list = g_list_append(bd->blog_list, blog);
		}

		blog->isAdmin = isAdmin;
		strcpy(blog->url, url ? url : "");
		g_free(url);
		strcpy(blog->blogid, blogid ? blogid : "");
		g_free(blogid);
		if(strcmp(blog->lastPublished, "")!=0) {
			if(strcmp(lastPublished, blog->lastPublished) > 0)
				blog_changed = TRUE;
		}
		strcpy(blog->lastPublished, lastPublished ? lastPublished : "");
		g_free(lastPublished);
		blog->showTitle = showTitle;
		strcpy(blog->blogName,  blogName ? blogName : "");
		g_free(blogName);

		/* Add new Blog to Buddy list */
		buddy = gaim_find_buddy(gc->account, blog->blogName);
		if(!buddy) {
			GaimGroup *g;
			buddy = gaim_buddy_new(gc->account, blog->blogName, NULL);
			if (!(g = gaim_find_group(gname))) {
				g = gaim_group_new(gname);
				gaim_blist_add_group(g, NULL);
			}
			gaim_blist_add_buddy(buddy, NULL, g, NULL);
		}
		/* Keep UserBlog info with buddy and vice versa */
		buddy->proto_data = (void*)blog;
		blog->buddy = buddy;

		if(GET_OPT(ops, GET_USERS_BLOGS_PERIODIC)) {
			/* compare with last published val */
			if(blog_changed) {
				/* update blogger status */
				gaim_debug(GAIM_DEBUG_INFO, "blogger", "Blogger Status Update..\n");
				SET_OPT(blog->blog_status, BLOGGER_CHANGED);
				/* Check and set last posted to see if its a new post */
				blogger_get_recent_posts(gc, blog->blogName, 1, 
							 GET_POSTS_CHECK_FOR_NEW_POST | GET_USERS_BLOGS_PERIODIC);
			}
		}
		else if(GET_OPT(ops, GET_USERS_BLOGS_STARTUP))
			blogger_get_recent_posts(gc, blog->blogName, 1, GET_POSTS_LAST_POSTED);

		gaim_debug(GAIM_DEBUG_INFO, "blogger", "Doing buddy update on: %s\n", buddy->name);
		gaim_blist_update_buddy_presence(buddy, TRUE);

		data=data->next;
	}/* end while */
	
	/* free xmlrpc response tree */
	xmlrpc_free_response(resp);
	g_free(gname);

	if(GET_OPT(ops, GET_USERS_BLOGS_STARTUP)) {
		gaim_connection_set_state(gc, GAIM_CONNECTED);
		serv_finish_login(gc);
	}
}/* end blogger_get_users_blogs_cb()*/

static void blogger_get_user_info_cb(gc, recvbuf) {}

static void edit_template_send_cb(GtkButton *button, struct gen_editor_data *ged) {
	struct template_data *td = ged->data;
	char* content;

	gaim_debug(GAIM_DEBUG_INFO, "blogger", "edit_template_send_cb\n");
	content = get_txtbuf_content(ged->txtbuf);
	blogger_set_template(td->gc, td->blog_name, content, ged);
}

static void edit_template_destroy_cb(GtkWidget *widget, struct gen_editor_data *ged) {
	gaim_debug(GAIM_DEBUG_INFO, "blogger", "edit_template_destroy_cb\n");
	/* template specific data */
	g_free(ged->data);
	g_free(ged);
}

static void blogger_get_template_cb(GaimConnection *gc, const char* recvbuf, UserBlog *blog) {
	struct xmlrpc_response *resp;
	char *ret;
	int ret_len;
	struct template_data *td;

	gaim_debug(GAIM_DEBUG_INFO, "blogger", "blogger_get_template_cb: for %s\n", blog->blogName);

	if(!(resp = blogger_process_response(gc, getTemplate, recvbuf, 0)))
		return;

	ret = (char*)SAFE_POINTER_3(XMLRPC_PARAM(SAFE_POINTER_2(resp,data)),value,data);
	CHECK_POINTER(blogger_get_template_cb, ret);
	ret_len = XMLRPC_PARAM(SAFE_POINTER_2(resp,data))->value->data_len;
	ret[ret_len]='\0';

	td = g_new0(struct template_data, 1);
	td->gc = gc;
	td->blog_name = blog->blogName;

	show_generic_editor(_("Blog Template Editor"), 
			    blog->blogName,
			    _("Update Template"), 
			    _("Cancel"),
			    edit_template_send_cb, 
			    window_close_cb,
			    edit_template_destroy_cb,
			    ret,
			    (void*)td,
			    0);

	/* free xmlrpc response tree */
	xmlrpc_free_response(resp);
}

static void blogger_set_template_cb(GaimConnection *gc, const char* recvbuf, void* cb_data) {
	struct gen_editor_data *ged = cb_data;
	struct xmlrpc_response *resp;
	char *ret;

	gaim_debug(GAIM_DEBUG_INFO, "blogger", "blogger_set_template_cb\n");

	if(!(resp = blogger_process_response(gc, setTemplate, recvbuf, 0)))
		return;

	ret = (char*)SAFE_POINTER_3(XMLRPC_PARAM(SAFE_POINTER_2(resp,data)),value,data);
	if(atoi(ret) == 1) {
		gaim_debug(GAIM_DEBUG_INFO, "blogger", "blogger_set_template_cb: Template set .. Success\n");
		/* Destroy editor window */
		gtk_widget_destroy(ged->win);
	}
	else {
		gaim_debug(GAIM_DEBUG_WARNING, "blogger", "blogger_set_template_cb: Return without error, but response unexpected: %s\n", ret);
		/* Display error dialog */
		gaim_notify_error(NULL, NULL, _("Blogger server returned an unexpected response."), "setTemplate Method");
	}

	xmlrpc_free_response(resp);
}

static void blogger_callback(gpointer data, gint source, GaimInputCondition cond) {
	GaimConnection *gc = data;
	BloggerData *bd = gc->proto_data;
	BloggerReq pend;
	UserBlog *pending_blog;
	PendingOps pend_ops;
	void *cb_data;

	gaim_debug(GAIM_DEBUG_INFO, "blogger", "blogger_callback\n");

	if (!g_list_find(gaim_connections_get_all(), gc)) {
		gaim_debug(GAIM_DEBUG_WARNING, "blogger", "blogger_callback: Error - Could not find gaim connection\n");
		close(source);
		return;
	}
		
	/* Keep track of pending info in case we have nested callbacks */
	pend = bd->pending;
	pending_blog = bd->pending_blog;
	pend_ops = bd->pending_ops;
	cb_data = bd->cb_data;

	if (cond & GAIM_INPUT_READ) {
		char* recvbuf = NULL;
		int recvlen = 0;
		char buffer[1024];
		int len=0;

		/* Receiving request response */
		while((len=read(source, &buffer, 1024)) != -1 && len > 0) {
			recvbuf = g_realloc(recvbuf, recvlen + len);
			memcpy(recvbuf+recvlen, &buffer, len);
			recvlen+=len;
		}

		/* Finish up .. We may have more blogger calls from the callbacks below */
		reset_pending(bd);
		gaim_input_remove(gc->inpa);
		close(source);

		if(len == 0)
                        gaim_debug(GAIM_DEBUG_WARNING, "blogger", "blogger_callback: Connection was gracefully closed\n");
		if(recvlen == 0) {
			gaim_debug(GAIM_DEBUG_WARNING, "blogger", "blogger_callback: Error reading\n");
		}
		else {
			char *buf, *errmsg = NULL;
			int len;

			recvbuf = g_realloc(recvbuf, recvlen + 1);
			recvbuf[recvlen]='\0';
			buf = xmlrpc_get_http_post_response(recvbuf, &len, &errmsg);
			if(!buf || errmsg) {
				if(GET_OPT(pend_ops, GET_USERS_BLOGS_STARTUP)) {
					if (!errmsg) {
						errmsg = g_strdup(_("Error parsing HTTP."));
					}
					gaim_connection_error(gc, errmsg);
				}
				g_free(recvbuf);
				if (errmsg) {
					g_free(errmsg);
				}
				return;
			}

			switch(pend) {
			case newPost:
				blogger_new_post_cb(gc, buf, pending_blog);
				break;
			case editPost:
				blogger_edit_post_cb(gc, buf, pending_blog, cb_data);
				break;
			case deletePost:
				blogger_delete_post_cb(gc, buf, cb_data);
				break;
			case getRecentPosts:
				blogger_get_recent_posts_cb(gc, buf, pending_blog, pend_ops);
				break;
			case getUsersBlogs:
				blogger_get_users_blogs_cb(gc, buf, pend_ops);
				break;
			case getUserInfo:
				blogger_get_user_info_cb(gc, buf);
				break;
			case getTemplate:
				blogger_get_template_cb(gc, buf, pending_blog);
				break;
			case setTemplate:
				blogger_set_template_cb(gc, buf, cb_data);
				break;
			case none:
			default:
				gaim_debug(GAIM_DEBUG_WARNING, "blogger", "blogger_callback: Error - Unkown pending blogger method\n");
				gaim_debug(GAIM_DEBUG_INFO, "blogger", "Server response dump:\n%s\n", buf);
			}/* end switch */
			g_free(recvbuf);
		}		
	}
	else {
		gaim_debug(GAIM_DEBUG_WARNING, "blogger", "blogger_callback: Error - Did not get read condition\n");
		reset_pending(bd);
		gaim_input_remove(gc->inpa);
		close(source);
	}

	gaim_debug(GAIM_DEBUG_INFO, "blogger", "blogger_callback: End blogger method %s\n", blogger_get_method_string(pend));
	
	/* If there are any queued blogger requests, and none pending.. fire away */
	if((bd->pending == none) && !g_queue_is_empty(bd->request_queue)) {
		BloggerRequest *req = g_queue_pop_tail(bd->request_queue);
		blogger_send_request(gc, req->pend, req->request, req->blog, req->options);
		g_free(req);
	}
}

/* Blogger Menu Callbacks */

static void blogger_open_url(GaimBlistNode *node, gpointer ignore) {
	GaimBuddy *buddy;
	GaimConnection *gc;
	BloggerData *bd;
	UserBlog *blog;

	g_return_if_fail(GAIM_BLIST_NODE_IS_BUDDY(node));

	buddy = (GaimBuddy *) node;
	gc = gaim_account_get_connection(buddy->account);
	bd = gc->proto_data;

	blog = blogger_find_blog(bd, buddy->name);

	if(blog && blog->url) {
		gaim_notify_uri(NULL, blog->url);
	}
	else
		gaim_debug(GAIM_DEBUG_WARNING, "blogger", "blogger_open_url: Couldn't get URL for Blog: %s\n", buddy->name);
}

static void blogger_edit_recent(GaimBlistNode *node, gpointer ignore) {
	GaimBuddy *buddy;
	GaimConnection *gc;
	
	g_return_if_fail(GAIM_BLIST_NODE_IS_BUDDY(node));

	buddy = (GaimBuddy *) node;
	gc = gaim_account_get_connection(buddy->account);

	blogger_get_recent_posts(gc, buddy->name, 10, GET_POSTS_FOR_EDIT);
}

static void blogger_edit_template(GaimBlistNode *node, gpointer ignore) {
	GaimBuddy *buddy;
	GaimConnection *gc;
	
	g_return_if_fail(GAIM_BLIST_NODE_IS_BUDDY(node));

	buddy = (GaimBuddy *) node;
	gc = gaim_account_get_connection(buddy->account);

	blogger_get_template(gc, buddy->name);
}

static void blogger_status_reset(GaimBlistNode *node, gpointer ignore) {
	GaimBuddy *buddy;
	UserBlog *blog;
	
	g_return_if_fail(GAIM_BLIST_NODE_IS_BUDDY(node));

	buddy = (GaimBuddy *) node;
	blog = buddy->proto_data;

	RESET_OPT(blog->blog_status);
	gaim_blist_update_buddy_status(buddy, buddy->uc);
}

static gboolean blogger_update_check_cb(gpointer data) {
	GaimConnection *gc = data;

	gaim_debug(GAIM_DEBUG_INFO, "blogger", "blogger_update_check_cb\n");

	/* Update Blog info */
	blogger_get_users_blogs(gc, GET_USERS_BLOGS_PERIODIC);
	return TRUE; /* don't remove timeout */
}


/**
 **  Exported Code - Registered Protocol Functions
 **/

static const char *blogger_list_icon(GaimAccount *a, GaimBuddy *b) {
	return "blogger";
}

static void blogger_login(GaimAccount *account) {
	GaimConnection *gc;
	BloggerData *bd;
	char buf[128];

	/* convert people using the old address over to the new one automatically 
	 * TODO: remove in 0.80 or so
	 */
	if (strcmp(gaim_account_get_string(account, "server", BLOGGER_HOST_URI), OLD_BLOGGER_HOST_URI) == 0) {
		gaim_debug(GAIM_DEBUG_INFO, "blogger", "converting from old URL ('%s') to new URL ('%s')\n", OLD_BLOGGER_HOST_URI, BLOGGER_HOST_URI);
		gaim_account_set_string(account, 
				"server", BLOGGER_HOST_URI);
	}

	gaim_debug(GAIM_DEBUG_INFO, "blogger", "blogger_login\n");

	gc = gaim_account_get_connection(account);
	gc->flags |= GAIM_CONNECTION_HTML;
	bd = gc->proto_data = g_new0(BloggerData, 1);
	bd->request_queue = g_queue_new();
	bd->gc = gc;

	/* Set startup so that blogger_get_users_cb knows we're signing on */
	SET_OPT(bd->pending_ops, GET_USERS_BLOGS_STARTUP);
	g_snprintf(buf, sizeof(buf), _("Connecting to Blogger Server."));
	gaim_connection_update_progress(gc, buf, 2, 5);

	blogger_get_users_blogs(gc, GET_USERS_BLOGS_STARTUP);
	/* If we're not pending there was an error */
	if(!bd->pending) {
		gaim_connection_error(gc, _("Blogger plugin error"));
	}

	/* add timeout to periodically update blogger blist info */
	bd->update_timeout = gaim_timeout_add(BLOGGER_TIMEOUT, blogger_update_check_cb, gc);
}

static void blogger_close(GaimConnection *gc) {
	BloggerData *bd = gc->proto_data;

	gaim_debug(GAIM_DEBUG_INFO, "blogger", "blogger_close\n");

	if (gc->inpa > 0)
		gaim_input_remove(gc->inpa);
	/* If waiting for proxy connect cb to be called.. */
	if(bd->send_data)
		g_free(bd->send_data);

	/* Check for queued Blogger requests */
	if(!g_queue_is_empty(bd->request_queue)) {
		BloggerRequest *req;
		gaim_debug(GAIM_DEBUG_INFO, "blogger", "%d queued blogger request%s %s outstanding. Cancelling..\n",
			     bd->request_queue->length,
			     ((bd->request_queue->length == 1) ? "" : "s"),
			     ((bd->request_queue->length == 1) ? "is" : "are"));
		while((req=g_queue_pop_tail(bd->request_queue))) {
			g_free(req->request);
			g_free(req);
		}
	}
	g_queue_free(bd->request_queue);

	gaim_timeout_remove(bd->update_timeout);

	/* Do this before we destroy blogger_data */
	blogger_remove_invalid_blogs(gc);

	/* Destroy blog_list */
	free_bloglist(bd->blog_list);
	g_free(bd);
}

static int blogger_send_im(GaimConnection *gc, const char *name, 
		const char *message, GaimConvImFlags imflags) {
	BloggerData *bd = gc->proto_data;

	gaim_debug(GAIM_DEBUG_INFO, "blogger", "blogger_send_im\n");

	/* Don't send a message while one is still pending, and don't queue it. 
	   It would be better if we could return -1 to indicate not to send and
	   to leave the message text in the convo text entry box.*/
	if(bd->pending == newPost) {
		gaim_notify_warning(NULL, NULL,
				    _("The previous blogger post has not yet completed. Your message was not sent."),
				    NULL);
		return 0;
	}
	blogger_new_post(gc, name, message);
	return 1;
}

static void blogger_add_buddy(GaimConnection *gc, GaimBuddy *buddy, GaimGroup *foo) {
	gaim_debug(GAIM_DEBUG_INFO, "blogger", "blogger_add_buddy\n");
}

static void blogger_remove_buddy(GaimConnection *gc, GaimBuddy *buddy, GaimGroup *group) {
	gaim_debug(GAIM_DEBUG_INFO, "blogger", "blogger_remove_buddy\n");
}

static GList *blogger_buddy_menu(GaimBlistNode *node) {
	GList *m = NULL;
	if (GAIM_BLIST_NODE_IS_BUDDY(node)) {
		GaimBlistNodeAction *act;

		gaim_debug(GAIM_DEBUG_INFO, "blogger", "blogger_buddy_menu\n");

		act = gaim_blist_node_action_new(_("Open Blog"),
				blogger_open_url, NULL);
		m = g_list_append(m, act);

		act = gaim_blist_node_action_new(_("Edit Recent Posts"),
				blogger_edit_recent, NULL);
		m = g_list_append(m, act);

		act = gaim_blist_node_action_new(_("Edit Blog Template"),
				blogger_edit_template, NULL);
		m = g_list_append(m, act);

		act = gaim_blist_node_action_new(_("Reset Blog Status"),
				blogger_status_reset, NULL);
		m = g_list_append(m, act);
	}

	return m;
}

static void blogger_convo_closed(GaimConnection *gc, const char *who) {
	gaim_debug(GAIM_DEBUG_INFO, "blogger", "blogger_convo_closed\n");
}

static char *blogger_tooltip_text(GaimBuddy *b) {
	UserBlog *ub=NULL;

	ub = (UserBlog*)SAFE_POINTER_2(b,proto_data);
	if(ub && strcmp(ub->blogName, b->name)==0) {
		char* text;
		time_t t;
		char* date=NULL;
		
		t = convert_iso8601(ub->lastPublished);
		date = ctime(&t);
		date[strlen(date)-1] = '\0';
		text = g_strdup_printf("\n<b>%s</b> %s"     /* Blog User */
				       "\n<b>%s</b> %s"     /* Admin */
				       "\n<b>%s</b> %s"     /* URL */
				       "\n<b>%s</b> %s"     /* Last Published */
				       "\n<b>%s</b> %s",    /* Blog Status */
				       _("Blog User:"), b->account->username,
				       _("Admin:"), ub->isAdmin ? "Yes" : "No",
				       _("URL:"), ub->url ? ub->url : "",
				       _("Last Published:"), date ? date : "",
				       _("Status:"), GET_OPT(ub->blog_status, BLOGGER_NEW_POST) ? "New post" :
				                           GET_OPT(ub->blog_status, BLOGGER_CHANGED) ? "Blog Update" : "None");
		return text;
	}

	CHECK_POINTER(blogger_tooltip_text,ub);

	return NULL;
}

static void blogger_list_emblems(GaimBuddy *b, char **se, char **sw, char **nw, char **ne) {
	UserBlog *blog = SAFE_POINTER_2(b,proto_data);

	if(blog && GET_OPT(blog->blog_status, BLOGGER_CHANGED))
		*se = "away";
	if(blog && GET_OPT(blog->blog_status, BLOGGER_NEW_POST))
		*se = "freeforchat";
}

static GaimPluginProtocolInfo prpl_info =
{
	0,			/* options */
	NULL,			/* user_splits */
	NULL,			/* protocol_options */
	NO_BUDDY_ICONS,		/* icon_spec */
	blogger_list_icon,	/* list_icon */
	blogger_list_emblems,	/* list_emblems */
	NULL,			/* status_text */
	blogger_tooltip_text,	/* tooltip_text */
	NULL,			/* away_states */
	blogger_buddy_menu,	/* blist_node_menu */
	NULL,			/* chat_info */
	NULL,			/* chat_info_defaults */
	blogger_login,		/* login */
	blogger_close,		/* close */
	blogger_send_im,	/* send_im */
	NULL,			/* set_info */
	NULL,			/* send_typing */
	NULL,			/* get_info */
	NULL,			/* set_away */
	NULL,			/* set_idle */
	NULL,			/* change_passwd */
	blogger_add_buddy,	/* add_buddy */
	NULL,			/* add_buddies */
	blogger_remove_buddy,	/* remove_buddy */
	NULL,			/* remove_buddies */
	NULL,			/* add_permit */
	NULL,			/* add_deny */
	NULL,			/* rem_permit */
	NULL,			/* rem_deny */
	NULL,			/* set_permit_deny */
	NULL,			/* warn */
	NULL,			/* join_chat */
	NULL,			/* reject_chat */
	NULL,			/* get_chat_name */
	NULL,			/* chat_invite */
	NULL,			/* chat_leave */
	NULL,			/* chat_whisper */
	NULL,			/* chat_send */
	NULL,			/* keepalive */
	NULL,			/* register_user */
	NULL,			/* get_cb_info */
	NULL,			/* get_cb_away */
	NULL,			/* alias_buddy */
	NULL,			/* group_buddy */
	NULL,			/* rename_group */
	NULL,			/* buddy_free */
	blogger_convo_closed,	/* convo_closed */
	NULL,			/* normalize */
	NULL,			/* set_buddy_icon */
	NULL,			/* remove_group */
	NULL,			/* get_cb_real_name */
	NULL,			/* set_chat_topic */
	NULL,			/* find_blist_chat */
	NULL,			/* roomlist_get_list */
	NULL,			/* roomlist_cancel */
	NULL			/* roomlist_expand_category */
};

static GaimPluginInfo info =
{
	GAIM_PLUGIN_MAGIC,
	GAIM_MAJOR_VERSION,
	GAIM_MINOR_VERSION,
	GAIM_PLUGIN_PROTOCOL,                             /**< type           */
	NULL,                                             /**< ui_requirement */
	0,                                                /**< flags          */
	NULL,                                             /**< dependencies   */
	GAIM_PRIORITY_DEFAULT,                            /**< priority       */

	"prpl-blogger",                                   /**< id             */
	"BLOGGER",                                        /**< name           */
	VERSION,                                          /**< version        */
	                                                  /**  summary        */
	N_("BLOGGER Protocol Plugin"),
	                                                  /**  description    */
	N_("BLOGGER Protocol Plugin"),
	"Herman Bloggs <hermanator12002@yahoo.com>",      /**< author         */
	"http://gaim-blogger.sourceforge.net",            /**< homepage       */

	NULL,                                             /**< load           */
	NULL,                                             /**< unload         */
	NULL,                                             /**< destroy        */

	NULL,                                             /**< ui_info        */
	&prpl_info,                                       /**< extra_info     */
	NULL,                                             /**< prefs_info     */
	NULL                                              /**< actions        */
};

static void
init_plugin(GaimPlugin *plugin)
{
	GaimAccountOption *option;

	option = gaim_account_option_string_new(_("Host URI"),
						"server",
						BLOGGER_HOST_URI);
	prpl_info.protocol_options = g_list_append(prpl_info.protocol_options, option);

	option = gaim_account_option_int_new(_("Port:"),
						"port",
						BLOGGER_HOST_PORT);
	prpl_info.protocol_options = g_list_append(prpl_info.protocol_options, option);

	my_protocol = plugin;

	calculate_server_tz_offset();
}

GAIM_INIT_PLUGIN(blogger, init_plugin, info);

